(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`AMPS`", {"ProcessOptions`", "AtomicDensityMatrix`"}];

Begin["`Private`"];

(*Angular momentum probability distribution*)

(*Angular momentum probability distribution in terms of the \[Kappa]-q tensor elements [Ale2004, Eq. (A10)]*)

\[Rho]JJ = MaxProjectionProbability;

ProjectionProbability[a_?TensorDecompositionQ, th_, ph_, crit_] := Module[
    {rank = Rank@a, cont = ToContravariant@a, j, population},
    j = Last@rank / 2;
    population = {Normalize[Boole@StateMatchQ[ZeemanSystem@j, crit], Total]};
    Sqrt[(4 Pi) / (2j + 1)] *
        Sum[
            Component[population, m] CG[{j, m}, {k, 0}, {j, m}] Component[cont, k, q] SphericalHarmonicY[k, q, th, ph],
            {k, 0, 2 j},
            {q, -k, k},
            {m, -j, j}
        ] /; MatchQ[rank, {0, ___}]
];

(*Angular momentum probability distribution in terms of the m m' tensor elements [Ale2004, Eq. (A12)]*)

ProjectionProbability[rho_?OperatorQ, th_, ph_, spec_] := Module[{dm},
    dm = If[
        MatchQ[TensorType@spec, {"Contravariant" | "Covariant" | "Cartesian"}] && Rank@spec == Rank@rho,
        Vee[ToContravariant@spec, ToCovariant@spec],
        PopulatedDM[Rank@rho, spec]
    ];
    Tr[rho.WignerRotate[dm, {ph, th, 0}]]
];

MaxProjectionProbability[rho_, th_, ph_] := ProjectionProbability[rho, th, ph, M == J];

Options[AMPSPlot] = Join[{ProjectionState -> M == J}, Options[SphericalPlot3D]];

SetOptions[AMPSPlot, Boxed -> False, Axes -> False, PlotRange -> All];

(AMPSPlot[rho_, opts : OptionsPattern[]] := Module[{projectionState, th, ph},
    SphericalPlot3D[
        Evaluate@Re@ProjectionProbability[rho, th, ph, projectionState],
        {th, 0, Pi},
        {ph, 0, 2 Pi},
        Evaluate@FilterRules@opts
    ]
]) // ProcessOptions;

End[];

EndPackage[];
