(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`PolarizationMoments`", {"ProcessOptions`", "AtomicDensityMatrix`", "AtomicDensityMatrix`Common`"}];

AtomicSystemSpecificationQ;

TensorRecompose::nosupport = "The specified representation does not support rank(s) `1` of the tensor of type `2`.";

Begin["`Private`"];

Options[PolarizationMoments] = Join[
    FilterRules[Options@DMVariables, Except@Representation],
    {MaxRanks -> Infinity}
];

OptionProcessor[MaxRanks] = (# /. Automatic :> OptionValue[PolarizationMoments, MaxRanks]&);
OptionValidator[MaxRanks] = _Integer?Positive | Infinity;

Options@ZeroTensorDecompositionMatrix = {MaxRanks -> Automatic};

ZeroTensorDecompositionMatrix[sys_?AtomicSystemQ, opts : OptionsPattern[]] := ZeroTensorDecompositionMatrix[sys, sys, opts];

(ZeroTensorDecompositionMatrix[sys1_?AtomicSystemQ, sys2_?AtomicSystemQ, OptionsPattern[]] := Module[
    {maxRanks, flist1 = F@RemoveZeemanSublevels@sys1, flist2 = F@RemoveZeemanSublevels@sys2},
    Table[
        Transpose@List@Table[0, {q, -k, k}],
        {f1, flist1},
        {f2, flist2},
        {k, Abs[f1 - f2], Min[f1 + f2, Abs[f1 - f2] + maxRanks - 1]}
    ]
]) // ProcessOptions;

Options@PolarizationMoment = Options@ZeroTensorDecompositionMatrix;

PolarizationMoment[sys_?AtomicSystemQ, All, {k_Integer, q : _Integer | {__Integer}}, opts : OptionsPattern[]] :=
    PolarizationMoment[sys, sys, All, {k, q}, opts];

PolarizationMoment[sys1_?AtomicSystemQ, sys2_?AtomicSystemQ, All, {k_Integer, q_Integer}, opts : OptionsPattern[]] :=
    PolarizationMoment[sys1, sys2, All, {k, {q}}, opts];

PolarizationMoment[sys1_?AtomicSystemQ, sys2_?AtomicSystemQ, All, {k_Integer, qs : {__Integer}}, opts : OptionsPattern[]] := Module[
    {zero = ZeroTensorDecompositionMatrix[sys1, sys2, opts], kPosition},
    kPosition = Position[Rank@zero, k];
    Table[ReplacePart[zero, {##, -q + k + 1, 1}& @@@ kPosition -> 1], {q, qs}]
];

Options@PolarizationMomentPosition = Options@ZeroTensorDecompositionMatrix;

(*PolarizationMomentPosition[sys_?AtomicSystemQ, crit1_ -> crit2_, n_ : Infinity] :=*)
(*Module[{pos1, pos2, levels = RemoveZeemanSublevels@sys},*)
(*{pos1, pos2} = Flatten@StatePosition[levels, #] & /@ {crit1, crit2};*)
(*Union[#, Reverse /@ #]&@Flatten[Outer[List, pos1, pos2], 1]*)
(*];*)
(*PolarizationMomentPosition[sys_?AtomicSystemQ, crit : {__Rule}] :=*)
(*Union @@ (MatrixElementPosition[sys, #] &) /@ crit;*)

(PolarizationMomentPosition[sys_?AtomicSystemQ, All | True, opts : OptionsPattern[]] :=
    Position[ZeroTensorDecompositionMatrix[sys, FilterRules@opts], 0]
) // ProcessOptions;
PolarizationMomentPosition[sys_?AtomicSystemQ, None | False, OptionsPattern[]] := {};

Options[ExpandPMVariables] = FilterRules[Options@DMVariables, DMSymbol | DMLabel | ComplexExpandVariables];

myAssociation[rules : {__Rule}] := Function[{key}, Switch[key, ##]]& @@ Flatten[rules, Infinity, Rule];

(ExpandPMVariables[sys_, expr_, OptionsPattern[]] := Module[{dmLabel, dmSymbol, complexExpandVariables, levels, p, f},
    If[complexExpandVariables === False,
        expr,
        levels = RemoveZeemanSublevels@sys;
        {p, f} = myAssociation@Thread[LevelLabel@levels -> #]& /@ {Range@Length@levels, F@levels};
        With[{
            patt = Subscript[dmSymbol, dmLabel, {l1_, l2_}, {k_, q_}],
            elem = Subscript[dmSymbol, #1, dmLabel, {l1, l2}, {k, q}, ##2]&,
            partPatt = Subscript[dmSymbol, r : Re | Im, dmLabel, {l1_, l2_}, {k_, q_}, PatternSequence[] | PatternSequence[t_] | PatternSequence[t_, d_]],
            partElem = Subscript[dmSymbol, r, dmLabel, {##}, {k, -q}, t, d]&},
            expr /. {
                Derivative[d_][patt][t_] -> elem[Re, t, d] + I elem[Im, t, d],
                patt[t_] -> elem[Re, t] + I elem[Im, t],
                patt -> elem[Re] + I elem[Im]
            } /. {
                partPatt /; r == Re && p@l1 > p@l2 :> (-1)^(f@l1 - f@l2 + q) partElem[l2, l1],
                partPatt /; r == Im && p@l1 > p@l2 :> -(-1)^(f@l1 - f@l2 + q) partElem[l2, l1],
                partPatt /; r == Re && p@l1 == p@l2 && q <= 0 :> (-1)^q partElem[l1, l2],
                partPatt /; r == Im && p@l1 == p@l2 && q <= 0 :> Sign@q(-1)^q partElem[l1, l2]
            } /. Switch[complexExpandVariables,
                True,
                partPatt -> r[Derivative[d][elem@Unevaluated[]][t]],
                Subscript,
                partPatt -> Derivative[d][elem@r][t]
            ] /. (s : patt | partPatt)[] :> s
        ]
    ]
]) // ProcessOptions;

PolarizationMoments[j_?HalfIntegerQ, opts : OptionsPattern[]] := PolarizationMoments[ZeemanSystem[j], opts];

PolarizationMoments[label_, j_?HalfIntegerQ, opts : OptionsPattern[]] := PolarizationMoments[ZeemanSystem[label, j], opts];

(PolarizationMoments[sys_?AtomicSystemQ, opts : OptionsPattern[]] := Module[
    {dmSymbol, dmLabel, timeVariable, timeDependence, allowedElements, neglectedElements,
        rank, rankList, elements, levels, labels, pos, zero, f},
    levels = RemoveZeemanSublevels@sys;
    labels = LevelLabel@levels;
    pos = Complement @@ (PolarizationMomentPosition[sys, #, FilterRules@opts]&) /@ {allowedElements, neglectedElements};
    f = F@levels;
    zero = ZeroTensorDecompositionMatrix[sys, FilterRules@opts];
    rank = Rank@zero;
    rankList = Extract[rank, Take[#, 3]]& /@ pos;
    elements = MapThread[Subscript[dmSymbol, dmLabel, labels[[Take[#1, 2]]], {#2, -#1[[4]] + #2 + 1}]&, {pos, rankList}];
    If[timeDependence, elements = Through@elements@timeVariable];
    ExpandPMVariables[sys, ReplacePart[zero, Thread[pos -> elements]], FilterRules@opts]
]) // ProcessOptions;

Options[PMVariables] = Thread[Options[PolarizationMoments][[All, 1]] -> Automatic];

(PMVariables[sys_?AtomicSystemQ, opts : OptionsPattern[]] := Module[{maxRanks},
    DMVariables[sys, Representation -> {"PolarizationMoments", MaxRanks -> maxRanks}, FilterRules@opts]
]) // ProcessOptions;

Options[PolarizationMatrices] = {MaxRanks -> Automatic};

PolarizationMatrices[j_?HalfIntegerQ, opts : OptionsPattern[]] := Component@PolarizationMatrices[ZeemanSystem@j, opts];

(PolarizationMatrices[sys_?AtomicSystemQ, opts : OptionsPattern[]] := Module[{r},
    ECT@PolarizationMoments[sys, TimeDependence -> False, DMSymbol -> r, ComplexExpandVariables -> False, DMLabel -> None, FilterRules@opts] /.
        {{___, Subscript[r, {s1_, s2_}, {k_, 0}], ___}} :> WignerEckart[sys, {Polarization[s1, s2], k}]
]) // ProcessOptions;

(* Convert between m m' and \[Kappa] q representation (contravariant components) using the polarization operator.
   [Var88, Sec. 6.1.5, Eqs. (49,50)], also [Ale2004, Eqs. (A3-8) ]*)

AtomicSystemSpecificationQ[expr_] := AtomicSystemQ[expr] || HalfIntegerQ[expr];

Options[Decomposition] = {MaxRanks -> Automatic};

Decomposition[sys_?AtomicSystemQ, a_?OperatorQ, OptionsPattern[]] /; AtomicSystemType@sys == "Toy" := a;

Decomposition[a_?OperatorQ, opts : OptionsPattern[]] := Decomposition[Rank@a, a, opts];

Decomposition[a_?TensorOperatorQ, opts : OptionsPattern[]] :=
    Decomposition[First@Flatten@ComponentMap[Rank, a, 1], a, opts];

(Decomposition[sys_?AtomicSystemSpecificationQ, a_, opts : OptionsPattern[]] :=
    TensorTrace[a, ECT@PolarizationMatrices[sys, FilterRules@opts]]
) // ProcessOptions;

Options[Recomposition] = {MaxRanks -> Automatic};

Recomposition[a_?TensorDecompositionQ, opts : OptionsPattern[]] := Module[{rank = Rank@a},
    Recomposition[Last@rank / 2, a, opts] /; MatchQ[rank, {0, ___}]
];

(Recomposition[sys_?AtomicSystemSpecificationQ, a_, opts : OptionsPattern[]] :=
    TensorRecompose[a, PolarizationMatrices[sys, FilterRules@opts]]
) // ProcessOptions;

TensorRecompose[
    {a_, {"TensorDecomposition"..., "Covariant"}},
    {p_, {"TensorDecomposition", "Covariant", "Operator"}}
] := CenterDot[a, ECT@Component[p, Rank@a]];

TensorRecompose[
    {a_, {"TensorDecomposition"..., "Contravariant"}},
    {p_, {"TensorDecomposition", "Covariant", "Operator"}}
] := CenterDot[Component[p, Rank@a], a];

TensorRecompose[
    {a_, {"TensorDecompositionMatrix", "TensorDecomposition", "Covariant"}},
    {p_, {"TensorDecompositionMatrix", "TensorDecomposition", "Covariant", "Operator"}}
] := CenterDot[a, ECT@p];

TensorRecompose[
    {a_, {"TensorDecompositionMatrix", "TensorDecomposition", "Contravariant"}},
    {p_, {"TensorDecompositionMatrix", "TensorDecomposition", "Covariant", "Operator"}}
] := CenterDot[p, a];

TensorRecompose[a : Except@TypedTensorPattern[], p : Except@TypedTensorPattern[]] := Module[
    {aType = TensorType@a, pType = TensorType@p, noMessage = True, result},
    result = Check[
        QuietCheck[
            TensorRecompose[{a, aType}, {p, pType}],
            Message[TensorRecompose::nosupport, Rank@a, aType],
            {Component::noexist}
        ],
        noMessage = False
    ];
    result /; noMessage && TensorTypeQ@aType && TensorTypeQ@pType
];

(*DecompositionProduct[sys_?AtomicSystemQ, a_?CovariantTensorDecompositionMatrixQ, b_?CovariantTensorDecompositionMatrixQ] :=*)
(*Module[{timeVariable, levels = RemoveZeemanSublevels@sys, pm, n, f, component},*)
(*rank = Rank@a;*)
(*f = f@levels;*)
(*n = Length@levels;*)
(*component = Function[{a, i, j, k, q}, op[[i, j, k - rank[[i, j, 1]] + 1, -q + k + 1, 1]]];*)
(*Flatten@Table[*)
(*Transpose@{Table[*)
(*Sum[(-1)^(2 f[[i1]] + k) component[a, i1, i2, k1, q1] component[b, i2, i3, k2, q2] **)
(*Sqrt[(2 k1 + 1) (2 k2 + 1)] CG[{k1, q1}, {k2, q2}, {k, q}] * SJS[{k1, k2, k}, {f[[i3]], f[[i1]], f[[i2]]}],*)
(*{i2, n},*)
(*{k1, rank[[i1, i2, 1]], rank[[i1, i2, 2]]},*)
(*{q1, k1, -k1, -1},*)
(*{k2, rank[[i2, i3, 1]], rank[[i2, i3, 2]]},*)
(*{q2, k2, -k2, -1}*)
(*],*)
(*{q, k, -k, -1}*)
(*]},*)
(*{i1, n},*)
(*{i3, n},*)
(*{k, rank[[i1, i3, 1]], rank[[i1, i3, 2]]}*)
(*]*)
(*];*)

End[];

EndPackage[];
